/*
Copyright (c) 2011, Tom Van Cutsem, Vrije Universiteit Brussel
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the Vrije Universiteit Brussel nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.*/

/**
 * A series of experiments with threads and locks.
 * 
 * The first few experiments use a shared integer counter that is incremented
 * concurrently by multiple threads.
 * 
 * The experiment {@link #runWithoutLock()} increments the counter without synchronization.
 * What will be the effect on the counter value?
 * 
 * The experiment {@link #runWithLock()} uses synchronization with locks to increment
 * the counter. What will the counter value be after the threads have finished updating?
 * 
 * The experiment {@link #runWithLockMalicious()} illustrates what happens when all n-1
 * threads acquire a lock, but the nth thread forgets to acquire the lock. What will the
 * counter value be after running the experiment?
 * 
 * The next set of experiments illustrates that critical sections do not automatically
 * compose into larger critical sections. Two threads share a counter object initialized
 * to an even integer. One thread always increments the counter twice, leading to a new
 * even number. Another thread reads the value of the counter and checks whether it is
 * odd.
 * 
 * The experiment {@link #evenExperimentWrong()} illustrates that the reader thread can
 * observe odd values if the two counter updates are not combined into a single critical
 * section. {@link #evenExperimentRight()} shows how to do the update right.
 * 
 * @author tvcutsem
 */
public class ThreadTest {

	private static final int MAX_THREADS = 10;
	private static final int NUM_INCS = 100000;
	public int val = 0;
	
	public void inc() { val = val + 1; }
	
	/**
	 * MAX_THREADS each increment a shared counter NUM_INCS times.
	 * Result: the counter's value is most likely not MAX_THREADS * NUM_INCS
	 * due to race conditions. 
	 */
	public static void runWithoutLock() {
		
		final ThreadTest t = new ThreadTest();
		
		long before = System.currentTimeMillis();
		
		Thread[] threads = new Thread[MAX_THREADS];
		for (int i = 0; i < MAX_THREADS; i++) {
			threads[i] = new Thread(new Runnable() {
				public void run() {
					for (int j = 0; j < NUM_INCS; j++) {
						t.inc();
					}
				}
			});
			threads[i].start();
		}
		
		// wait for all threads to finish
		for (int j = 0; j < threads.length; j++) {
			try {
				threads[j].join();
			} catch (InterruptedException e) {}
		}
		long after = System.currentTimeMillis();
		
		System.out.println("runWithoutLock:");
		System.out.println("times incremented = "+ NUM_INCS * MAX_THREADS);
		System.out.println("c = "+ t.val);
		System.out.println("Running time (ms): " + (after - before));
	}
	
	/**
	 * MAX_THREADS each increment a shared synchronized counter NUM_INCS times.
	 * Result: the counter's value is MAX_THREADS * NUM_INCS as expected.
	 */
	public static void runWithLock() {
		
		final ThreadTest t = new ThreadTest();
		
		long before = System.currentTimeMillis();
		
		Thread[] threads = new Thread[MAX_THREADS];
		for (int i = 0; i < MAX_THREADS; i++) {
			threads[i] = new Thread(new Runnable() {
				public void run() {
					for (int j = 0; j < NUM_INCS; j++) {
						synchronized(t) { t.inc(); }						
					}
				}
			});
			threads[i].start();
		}
		
		// wait for all threads to finish
		for (int j = 0; j < threads.length; j++) {
			try {
				threads[j].join();
			} catch (InterruptedException e) {}
		}
		long after = System.currentTimeMillis();
		
		System.out.println("runWithLock:");
		System.out.println("times incremented = "+ NUM_INCS * MAX_THREADS);
		System.out.println("c = "+ t.val);
		System.out.println("Running time (ms): " + (after - before));
	}
	
	/**
	 * MAX_THREADS - 1 each increment a shared synchronized counter NUM_INCS times.
	 * One "malicious" thread increments the shared counter without synchronization.
	 * Result: the counter's value is most likely not MAX_THREADS * NUM_INCS
	 * due to race conditions. 
	 */
	public static void runWithLockMalicious() {
		
		final ThreadTest t = new ThreadTest();
		
		long before = System.currentTimeMillis();
		
		Thread[] threads = new Thread[MAX_THREADS - 1];
		for (int i = 0; i < threads.length; i++) {
			threads[i] = new Thread(new Runnable() {
				public void run() {
					for (int j = 0; j < NUM_INCS; j++) {
						synchronized(t) { t.inc(); }						
					}
				}
			});
			threads[i].start();
		}
		
		// malicious thread does not take the lock on the counter
		Thread malicious = new Thread(new Runnable() {
			public void run() {
				for (int j = 0; j < NUM_INCS; j++) {
					t.inc();					
				}
			}
		});
		malicious.start();
		
		// wait for all threads to finish
		for (int j = 0; j < threads.length; j++) {
			try {
				threads[j].join();
			} catch (InterruptedException e) {}
		}
		
		// wait for the malicious thread to finish
		try {
			malicious.join();
		} catch (InterruptedException e) {}
		
		long after = System.currentTimeMillis();
		
		System.out.println("runWithLockMalicious:");
		System.out.println("times incremented = "+ NUM_INCS * MAX_THREADS);
		System.out.println("c = "+ t.val);
		System.out.println("Running time (ms): " + (after - before));
	}
	
	private static class Counter {
      private int val = 0;
	  public synchronized void inc() { val = val + 1; }
	  public synchronized int count() { return val; }
	}
	
	/**
	 * One thread continuously increments the counter twice, without synchronization.
	 * Another thread reads the counter values in parallel and tests whether it ever
	 * sees an odd value.
	 * Result: the inspector thread will most likely see an odd value because of
	 * interleaving.
	 */
	public static void evenExperimentWrong() {
		final Counter c = new Counter();
		
		Thread evenCounter = new Thread(new Runnable() {
			  public void run() {
			    for (int j = 0; j < NUM_INCS; j++) {
			      c.inc();
			      c.inc();
			    }
			  }
			});
		evenCounter.start();
		
		Thread inspector = new Thread(new Runnable() {
			  boolean sawOdd = false;
			  public void run() {
			    for (int j = 0; j < NUM_INCS; j++) {
			      sawOdd = sawOdd | (c.count() % 2 == 1);
			    }
			    System.out.println("evenExperimentWrong: saw odd? " + sawOdd);
			  }
		});
		inspector.start();
	}
	
	/**
	 * One thread continuously increments the counter twice, with synchronization.
	 * Another thread reads the counter values in parallel and tests whether it ever
	 * sees an odd value.
	 * Result: the inspector thread will not detect an odd value.
	 */
	public static void evenExperimentRight() {
		final Counter c = new Counter();
		
		Thread evenCounter = new Thread(new Runnable() {
			  public void run() {
			    for (int j = 0; j < NUM_INCS; j++) {
			      synchronized (c) {
			    	  c.inc();
				      c.inc();
			      }
			    }
			  }
			});
		evenCounter.start();
		
		Thread inspector = new Thread(new Runnable() {
			  boolean sawOdd = false;
			  public void run() {
			    for (int j = 0; j < NUM_INCS; j++) {
			      sawOdd = sawOdd | (c.count() % 2 == 1);
			    }
			    System.out.println("evenExperimentRight: saw odd? " + sawOdd);
			  }
		});
		inspector.start();
	}
			
	public static void main(String[] args) {
		runWithoutLock();
		runWithLock();
		runWithLockMalicious();
		evenExperimentWrong();
		evenExperimentRight();
	}
	
}
